import numpy as np
from scipy.special import ellipk, ellipe

from ..base import BaseDipole, BaseMagneticDipole, BaseEM
from ... import spatial

__all__ = [
    "MagneticDipoleWholeSpace", "CircularLoopWholeSpace",
    "MagneticPoleWholeSpace"
]


class MagneticDipoleWholeSpace(BaseEM, BaseMagneticDipole):
    """Class for a static magnetic dipole in a wholespace.

    The ``MagneticDipoleWholeSpace`` class is used to analytically compute the
    fields and potentials within a wholespace due to a static magnetic dipole.

    """

    def vector_potential(self, xyz, coordinates="cartesian"):
        r"""Compute the vector potential for the static magnetic dipole.

        This method computes the vector potential for the magnetic dipole at
        the set of gridded xyz locations provided. Where :math:`\mu` is the
        magnetic permeability, :math:`\mathbf{m}` is the dipole moment,
        :math:`\mathbf{r_0}` the dipole location and :math:`\mathbf{r}`
        is the location at which we want to evaluate
        the vector potential :math:`\mathbf{a}`:

        .. math::

            \mathbf{a}(\mathbf{r}) = \frac{\mu}{4\pi}
            \frac{\mathbf{m} \times \, \Delta \mathbf{r}}{| \Delta r |^3}

        where

        .. math::
            \mathbf{\Delta r} = \mathbf{r} - \mathbf{r_0}

        For reference, see equation 5.83 in Griffiths (1999).

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we are calculating the vector potential
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic vector potential at each observation location in the
            coordinate system specified in units *Tm*.

        Examples
        --------
        Here, we define a z-oriented magnetic dipole and plot the vector
        potential on the xy-plane that intercepts at z=0.

        >>> from geoana.em.static import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the magnetic dipole.

        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> dipole_object = MagneticDipoleWholeSpace(
        >>>     location=location, orientation=orientation, moment=moment
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.linspace(-1, 1, 20), np.array([0]))
        >>> a = dipole_object.vector_potential(xyz)

        Finally, we plot the vector potential on the plane. Given the symmetry,
        there are only horizontal components.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0:2], a[:, 0:2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Vector potential at z=0')

        """
        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        n_obs = xyz.shape[0]

        # orientation of the dipole
        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        dxyz = self.vector_distance(xyz)
        r = spatial.repeat_scalar(self.distance(xyz))
        m = self.moment * np.atleast_2d(self.orientation).repeat(n_obs, axis=0)

        m_cross_r = np.cross(m, dxyz)
        a = (self.mu / (4 * np.pi)) * m_cross_r / (r**3)

        if coordinates.lower() == "cylindrical":
            a = spatial.cartesian_2_cylindrical(xyz, a)

        return a

    def magnetic_flux_density(self, xyz, coordinates="cartesian"):
        r"""Compute magnetic flux density produced by the static magnetic dipole.

        This method computes the magnetic flux density produced by the static magnetic
        dipole at gridded xyz locations provided. Where :math:`\mu` is the magnetic
        permeability of the wholespace, :math:`\mathbf{m}` is the dipole moment,
        :math:`\mathbf{r_0}` the dipole location and :math:`\mathbf{r}` is the location
        at which we want to evaluate the magnetic flux density :math:`\mathbf{B}`:

        .. math::

            \mathbf{B}(\mathbf{r}) = \frac{\mu}{4\pi} \Bigg [
            \frac{3 \Delta \mathbf{r} \big ( \mathbf{m} \cdot \, \Delta \mathbf{r} \big ) }{| \Delta \mathbf{r} |^5}
            - \frac{\mathbf{m}}{| \Delta \mathbf{r} |^3} \Bigg ]

        where

        .. math::
            \mathbf{\Delta r} = \mathbf{r} - \mathbf{r_0}

        For reference, see equation Griffiths (1999).

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray
            gridded locations at which we calculate the magnetic flux density
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic flux density at each observation location in the
            coordinate system specified in Teslas.


        Examples
        --------
        Here, we define a z-oriented magnetic dipole and plot the magnetic
        flux density on the xy-plane that intercepts y=0.

        >>> from geoana.em.static import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the magnetic dipole.

        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> dipole_object = MagneticDipoleWholeSpace(
        >>>     location=location, orientation=orientation, moment=moment
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.array([0]), np.linspace(-1, 1, 20))
        >>> B = dipole_object.magnetic_flux_density(xyz)

        Finally, we plot the vector potential on the plane. Given the symmetry,
        there are only horizontal components.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], B[:, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Magnetic flux density at y=0')

        """

        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        n_obs = xyz.shape[0]

        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        r = self.vector_distance(xyz)
        dxyz = spatial.repeat_scalar(self.distance(xyz))
        m_vec = (
            self.moment * np.atleast_2d(self.orientation).repeat(n_obs, axis=0)
        )

        m_dot_r = (m_vec * r).sum(axis=1)

        # Repeat the scalars
        m_dot_r = np.atleast_2d(m_dot_r).T.repeat(3, axis=1)
        # dxyz = np.atleast_2d(dxyz).T.repeat(3, axis=1)

        b = (self.mu / (4 * np.pi)) * (
            (3.0 * r * m_dot_r / (dxyz ** 5)) -
            m_vec / (dxyz ** 3)
        )

        if coordinates.lower() == "cylindrical":
            b = spatial.cartesian_2_cylindrical(xyz, b)

        return b

    def magnetic_field(self, xyz, coordinates="cartesian"):
        r"""Compute the magnetic field produced by a static magnetic dipole.

        This method computes the magnetic field produced by the static magnetic dipole at
        the set of gridded xyz locations provided. Where :math:`\mathbf{m}` is the dipole
        moment, :math:`\mathbf{r_0}` is the dipole location and :math:`\mathbf{r}` is the
        location at which we want to evaluate the magnetic field :math:`\mathbf{H}`:

        .. math::

            \mathbf{H}(\mathbf{r}) = \frac{1}{4\pi} \Bigg [
            \frac{3 \Delta \mathbf{r} \big ( \mathbf{m} \cdot \, \Delta \mathbf{r} \big ) }{| \Delta \mathbf{r} |^5}
            - \frac{\mathbf{m}}{| \Delta \mathbf{r} |^3} \Bigg ]

        where

        .. math::
            \mathbf{\Delta r} = \mathbf{r} - \mathbf{r_0}

        For reference, see equation Griffiths (1999).

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we calculate the magnetic field
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic field at each observation location in the
            coordinate system specified in units A/m.

        Examples
        --------
        Here, we define a z-oriented magnetic dipole and plot the magnetic
        field on the xz-plane that intercepts y=0.

        >>> from geoana.em.static import MagneticDipoleWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the magnetic dipole.

        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> moment = 1.
        >>> dipole_object = MagneticDipoleWholeSpace(
        >>>     location=location, orientation=orientation, moment=moment
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-1, 1, 20), np.array([0]), np.linspace(-1, 1, 20))
        >>> H = dipole_object.magnetic_field(xyz)

        Finally, we plot the vector potential on the plane. Given the symmetry,
        there are only horizontal components.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], H[:, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Magnetic field at y=0')

        """
        return self.magnetic_flux_density(xyz, coordinates=coordinates) / self.mu


class MagneticPoleWholeSpace(BaseEM, BaseMagneticDipole):
    """Class for a static magnetic pole in a wholespace.

    The ``MagneticPoleWholeSpace`` class is used to analytically compute the
    fields and potentials within a wholespace due to a static magnetic pole.
    """

    def magnetic_flux_density(self, xyz, coordinates="cartesian"):
        r"""Compute the magnetic flux density produced by the static magnetic pole.

        This method computes the magnetic flux density produced by the static magnetic pole
        at the set of gridded xyz locations provided. Where :math:`\mu` is the magnetic
        permeability of the wholespace, :math:`m` is the moment amplitude,
        :math:`\mathbf{r_0}` the pole's location and :math:`\mathbf{r}` is the location
        at which we want to evaluate the magnetic flux density :math:`\mathbf{B}`:

        .. math::

            \mathbf{B}(\mathbf{r}) = \frac{\mu m}{4\pi} \frac{\Delta \mathbf{r}}{| \Delta \mathbf{r}|^3}

        where

        .. math::
            \mathbf{\Delta r} = \mathbf{r} - \mathbf{r_0}

        For reference, see equation Griffiths (1999).

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded xyz locations at which we calculate the magnetic flux density
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic flux density at each observation location in the
            coordinate system specified in units T.

        """

        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        n_obs = xyz.shape[0]

        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        r = self.vector_distance(xyz)
        dxyz = spatial.repeat_scalar(self.distance(xyz))

        b = self.moment * self.mu / (4 * np.pi * (dxyz ** 3)) * r

        if coordinates.lower() == "cylindrical":
            b = spatial.cartesian_2_cylindrical(xyz, b)

        return b

    def magnetic_field(self, xyz, coordinates="cartesian"):
        r"""Compute the magnetic field produced by the static magnetic pole.

        This method computes the magnetic field produced by the static magnetic pole at
        the set of gridded xyz locations provided. Where :math:`\mu` is the magnetic
        permeability of the wholespace, :math:`m` is the moment amplitude,
        :math:`\mathbf{r_0}` the pole's location and :math:`\mathbf{r}` is the location
        at which we want to evaluate the magnetic field :math:`\mathbf{H}`:

        .. math::

            \mathbf{G}(\mathbf{r}) = \frac{m}{4\pi} \frac{\Delta \mathbf{r}}{| \Delta \mathbf{r}|^3}

        where

        .. math::
            \mathbf{\Delta r} = \mathbf{r} - \mathbf{r_0}

        For reference, see equation Griffiths (1999).

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we calculate the magnetic field
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic field at each observation location in the
            coordinate system specified in units A/m.

        """
        return self.magnetic_flux_density(xyz, coordinates=coordinates) / self.mu


class CircularLoopWholeSpace(BaseEM, BaseDipole):
    """Class for a circular loop of static current in a wholespace.

    The ``CircularLoopWholeSpace`` class is used to analytically compute the
    fields and potentials within a wholespace due to a circular loop carrying
    static current.

    Parameters
    ----------
    current : float
        Electrical current in the loop (A). Default is 1.
    radius : float
        Radius of the loop (m). Default is :math:`\\pi^{-1/2}` so that the loop
        has a default dipole moment of 1 :math:`A/m^2`.
    """

    def __init__(self, radius=np.sqrt(1.0/np.pi), current=1.0, **kwargs):

        self.current = current
        self.radius = radius
        super().__init__(**kwargs)


    @property
    def current(self):
        """Current in the loop in Amps

        Returns
        -------
        float
            Current in the loop Amps
        """
        return self._current

    @current.setter
    def current(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"current must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("current must be greater than 0")

        self._current = value


    @property
    def radius(self):
        """Radius of the loop in meters

        Returns
        -------
        float
            Radius of the loop in meters
        """
        return self._radius

    @radius.setter
    def radius(self, value):

        try:
            value = float(value)
        except:
            raise TypeError(f"radius must be a number, got {type(value)}")

        if value <= 0.0:
            raise ValueError("radius must be greater than 0")

        self._radius = value

    def vector_potential(self, xyz, coordinates="cartesian"):
        r"""Compute the vector potential for the static loop in a wholespace.

        This method computes the vector potential for the cirular current loop
        at the set of gridded xyz locations provided. Where :math:`\mu` is the magnetic
        permeability, :math:`I d\mathbf{s}` represents an infinitessimal segment
        of current at location :math:`\mathbf{r_s}` and :math:`\mathbf{r}` is the location
        at which we want to evaluate the vector potential :math:`\mathbf{a}`:

        .. math::

            \mathbf{a}(\mathbf{r}) = \frac{\mu I}{4\pi} \oint
            \frac{1}{|\mathbf{r} - \mathbf{r_s}|} d\mathbf{s}


        The above expression can be solve analytically by using the appropriate
        change of coordinate transforms and the solution for a horizontal current
        loop. For a horizontal current loop centered at (0,0,0), the solution in
        radial coordinates is given by:

        .. math::

            a_\theta (\rho, z) = \frac{\mu_0 I}{\pi k}
            \sqrt{ \frac{R}{\rho^2}} \bigg [ (1 - k^2/2) \, K(k^2) - K(k^2) \bigg ]

        where

        .. math::

            k^2 = \frac{4 R \rho}{(R + \rho)^2 + z^2}

        and

        - :math:`\rho = \sqrt{x^2 + y^2}` is the horizontal distance to the test point
        - :math:`I` is the current through the loop
        - :math:`R` is the radius of the loop
        - :math:`E(k^2)` and :math:`K(k^2)` are the complete elliptic integrals

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we calculate the vector potential
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic vector potential at each observation location in the
            coordinate system specified in units *Tm*.

        Examples
        --------
        Here, we define a horizontal loop and plot the vector
        potential on the xy-plane that intercepts at z=0.

        >>> from geoana.em.static import CircularLoopWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the loop.

        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> radius = 0.5
        >>> simulation = CircularLoopWholeSpace(
        >>>     location=location, orientation=orientation, radius=radius
        >>> )

        Now we create a set of gridded locations and compute the vector potential.

        >>> xyz = ndgrid(np.linspace(-1, 1, 50), np.linspace(-1, 1, 50), np.array([0]))
        >>> a = simulation.vector_potential(xyz)

        Finally, we plot the vector potential on the plane. Given the symmetry,
        there are only horizontal components.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0:2], a[:, 0:2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_title('Vector potential at z=0')

        """

        eps = 1e-10
        supported_coordinates = ["cartesian", "cylindrical"]
        assert coordinates.lower() in supported_coordinates, (
            "coordinates must be in {}, the coordinate system "
            "you provided, {}, is not yet supported".format(
                supported_coordinates, coordinates
            )
        )

        # convert coordinates if not cartesian
        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)

        xyz = spatial.rotate_points_from_normals(
            xyz, np.array(self.orientation),  # work around for a properties issue
            np.r_[0., 0., 1.], x0=np.array(self.location)
        )

        n_obs = xyz.shape[0]
        dxyz = self.vector_distance(xyz)
        r = self.distance(xyz)

        rho = np.sqrt((dxyz[:, :2]**2).sum(1))

        k2 = (4 * self.radius * rho) / ((self.radius + rho)**2 +dxyz[:, 2]**2)
        k2[k2 > 1.] = 1.  # if there are any rounding errors

        E = ellipe(k2)
        K = ellipk(k2)

        # singular if rho = 0, k2 = 1
        ind = (rho > eps) & (k2 < 1)

        Atheta = np.zeros_like(r)
        Atheta[ind] = (
            (self.mu * self.current) / (np.pi * np.sqrt(k2[ind])) *
            np.sqrt(self.radius / rho[ind]) *
            ((1. - k2[ind] / 2.)*K[ind] - E[ind])
        )

        # assume that the z-axis aligns with the polar axis
        A = np.zeros_like(xyz)
        A[ind, 0] = Atheta[ind] * (-dxyz[ind, 1] / rho[ind])
        A[ind, 1] = Atheta[ind] * (dxyz[ind, 0] / rho[ind])

        # rotate the points to aligned with the normal to the source
        A = spatial.rotate_points_from_normals(
            A, np.r_[0., 0., 1.], np.array(self.orientation),
            x0=np.array(self.location)
        )

        if coordinates.lower() == "cylindrical":
            A = spatial.cartesian_2_cylindrical(xyz, A)

        return A

    def magnetic_flux_density(self, xyz, coordinates="cartesian"):
        r"""Compute the magnetic flux density for the current loop in a wholespace.

        This method computes the magnetic flux density for the cirular current loop
        at the set of gridded xyz locations provided. Where :math:`\mu` is the magnetic
        permeability, :math:`I d\mathbf{s}` represents an infinitessimal segment
        of current at location :math:`\mathbf{r_s}` and :math:`\mathbf{r}` is the location
        at which we want to evaluate the magnetic flux density :math:`\mathbf{B}`:

        .. math::

            \mathbf{B}(\mathbf{r}) = - \frac{\mu I}{4\pi} \oint
            \frac{(\mathbf{r}-\mathbf{r_s}) \times d\mathbf{s}}{|\mathbf{r} - \mathbf{r_0}|^3}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we calculate the magnetic flux density
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic flux density at each observation location in the
            coordinate system specified in units *T*.

        Examples
        --------
        Here, we define a horizontal loop and plot the magnetic flux
        density on the xz-plane that intercepts at y=0.

        >>> from geoana.em.static import CircularLoopWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the loop.

        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> radius = 0.5
        >>> simulation = CircularLoopWholeSpace(
        >>>     location=location, orientation=orientation, radius=radius
        >>> )

        Now we create a set of gridded locations and compute the magnetic flux density.

        >>> xyz = ndgrid(np.linspace(-1, 1, 50), np.array([0]), np.linspace(-1, 1, 50))
        >>> B = simulation.magnetic_flux_density(xyz)

        Finally, we plot the magnetic flux density on the plane.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], B[:, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Y')
        >>> ax.set_title('Magnetic flux density at y=0')

        """
        xyz = np.atleast_2d(xyz)
        # convert coordinates if not cartesian
        if coordinates.lower() == "cylindrical":
            xyz = spatial.cylindrical_2_cartesian(xyz)
        elif coordinates.lower() != "cartesian":
            raise TypeError(
                f"coordinates must be 'cartesian' or 'cylindrical', the coordinate "
                f"system you provided, {coordinates}, is not yet supported."
            )

        xyz = spatial.rotate_points_from_normals(
            xyz, np.array(self.orientation),  # work around for a properties issue
            np.r_[0., 0., 1.], x0=np.array(self.location)
        )
        # rotate all the points such that the orientation is directly vertical

        dxyz = self.vector_distance(xyz)
        r = self.distance(xyz)

        rho = np.linalg.norm(dxyz[:, :2], axis=-1)

        B = np.zeros((len(rho), 3))

        # for On axis points
        inds_axial = rho==0.0

        B[inds_axial, -1] = self.mu * self.current * self.radius**2 / (
            2 * (self.radius**2 + dxyz[inds_axial, 2]**2)**(1.5)
        )

        # Off axis
        alpha = rho[~inds_axial]/self.radius
        beta = dxyz[~inds_axial, 2]/self.radius
        gamma = dxyz[~inds_axial, 2]/rho[~inds_axial]

        Q = ((1+alpha)**2 + beta**2)
        k2 =  4 * alpha/Q

        # axial part:
        B[~inds_axial, -1] = self.mu * self.current / (2 * self.radius * np.pi * np.sqrt(Q)) * (
            ellipe(k2)*(1 - alpha**2 - beta**2)/(Q  - 4 * alpha) + ellipk(k2)
        )

        # radial part:
        B_rad = self.mu * self.current * gamma / (2 * self.radius * np.pi * np.sqrt(Q)) * (
            ellipe(k2)*(1 + alpha**2 + beta**2)/(Q  - 4 * alpha) - ellipk(k2)
        )

        # convert radial component to x and y..
        B[~inds_axial, 0] = B_rad * (dxyz[~inds_axial, 0]/rho[~inds_axial])
        B[~inds_axial, 1] = B_rad * (dxyz[~inds_axial, 1]/rho[~inds_axial])

        # rotate the vectors to be aligned with the normal to the source
        B = spatial.rotate_points_from_normals(
           B, np.r_[0., 0., 1.], np.array(self.orientation),
        )

        if coordinates.lower() == "cylindrical":
            B = spatial.cartesian_2_cylindrical(xyz, B)

        return B

    def magnetic_field(self, xyz, coordinates="cartesian"):
        r"""Compute the magnetic field for the current loop in a wholespace.

        This method computes the magnetic field for the cirular current loop
        at the set of gridded xyz locations provided. Where :math:`\mu` is the magnetic
        permeability, :math:`I d\mathbf{s}` represents an infinitessimal segment
        of current at location :math:`\mathbf{r_s}` and :math:`\mathbf{r}` is the location
        at which we want to evaluate the magnetic field :math:`\mathbf{H}`:

        .. math::

            \mathbf{H}(\mathbf{r}) = - \frac{I}{4\pi} \oint
            \frac{(\mathbf{r}-\mathbf{r_s}) \times d\mathbf{s}}{|\mathbf{r} - \mathbf{r_0}|^3}

        Parameters
        ----------
        xyz : (n, 3) numpy.ndarray xyz
            gridded locations at which we calculate the magnetic field
        coordinates: str {'cartesian', 'cylindrical'}
            coordinate system that the location (xyz) are provided.
            The solution is also returned in this coordinate system.
            Default: `"cartesian"`

        Returns
        -------
        (n, 3) numpy.ndarray
            The magnetic field at each observation location in the
            coordinate system specified in units A/m.

        Examples
        --------
        Here, we define a horizontal loop and plot the magnetic field
        on the xz-plane that intercepts at y=0.

        >>> from geoana.em.static import CircularLoopWholeSpace
        >>> from geoana.utils import ndgrid
        >>> from geoana.plotting_utils import plot2Ddata
        >>> import numpy as np
        >>> import matplotlib.pyplot as plt

        Let us begin by defining the loop.

        >>> location = np.r_[0., 0., 0.]
        >>> orientation = np.r_[0., 0., 1.]
        >>> radius = 0.5
        >>> simulation = CircularLoopWholeSpace(
        >>>     location=location, orientation=orientation, radius=radius
        >>> )

        Now we create a set of gridded locations and compute the magnetic field.

        >>> xyz = ndgrid(np.linspace(-1, 1, 50), np.array([0]), np.linspace(-1, 1, 50))
        >>> H = simulation.magnetic_field(xyz)

        Finally, we plot the magnetic field on the plane.

        >>> fig = plt.figure(figsize=(4, 4))
        >>> ax = fig.add_axes([0.15, 0.15, 0.8, 0.8])
        >>> plot2Ddata(xyz[:, 0::2], H[:, 0::2], ax=ax, vec=True, scale='log')
        >>> ax.set_xlabel('X')
        >>> ax.set_ylabel('Z')
        >>> ax.set_title('Magnetic field at y=0')

        """
        return self.magnetic_flux_density(xyz, coordinates=coordinates) / self.mu



"""
=================================================
Basic Utilities (:mod:`geoana.utils`)
=================================================
.. currentmodule:: geoana.utils

The ``geoana.utils`` module provides a set of rudimentary functions that are
commonly used within the code base.

.. autosummary::
  :toctree: generated/

  mkvc
  ndgrid
  check_xyz_dim

"""
import numpy as np

def mkvc(x, n_dims=1):
    """Creates a vector with specified dimensionality.

    This function converts a :class:`numpy.ndarray` to a vector. In general,
    the output vector has a dimension of 1. However, the dimensionality
    can be specified if the user intends to carry out a dot product with
    a higher order array.

    Parameters
    ----------
    x : array_like
        An array that will be reorganized and output as a vector. The input array
        will be flattened on input in Fortran order.
    n_dims : int
        The dimension of the output vector. :data:`numpy.newaxis` are appened to the
        output array until it has this many axes.

    Returns
    -------
    numpy.ndarray
        The output vector, with at least ``n_dims`` axes.

    Examples
    --------
    Here, we reorganize a simple 2D array as a vector and demonstrate the
    impact of the *n_dim* argument.

    >>> from geoana.utils import mkvc
    >>> import numpy as np

    >>> a = np.random.rand(3, 2)
    >>> a
    array([[0.33534155, 0.25334363],
           [0.07147884, 0.81080958],
           [0.85892774, 0.74357806]])

    >>> v = mkvc(a)
    >>> v
    array([0.33534155, 0.07147884, 0.85892774, 0.25334363, 0.81080958,
           0.74357806])

    In Higher dimensions:

    >>> for ii in range(1, 4):
    ...     v = mkvc(a, ii)
    ...     print('Shape of output with n_dim =', ii, ': ', v.shape)
    Shape of output with n_dim = 1 :  (6,)
    Shape of output with n_dim = 2 :  (6, 1)
    Shape of output with n_dim = 3 :  (6, 1, 1)
    """
    if isinstance(x, np.matrix):
        x = np.array(x)

    if hasattr(x, 'tovec'):
        x = x.tovec()

    assert isinstance(x, np.ndarray), "Vector must be a numpy array"

    if n_dims == 1:
        return x.flatten(order='F')
    elif n_dims == 2:
        return x.flatten(order='F')[:, np.newaxis]
    elif n_dims == 3:
        return x.flatten(order='F')[:, np.newaxis, np.newaxis]


def ndgrid(*args, **kwargs):
    """
    Form tensorial grid for 1, 2, or 3 dimensions.

    Returns as column vectors by default.

    To return as matrix input:

        ndgrid(..., vector=False)

    The inputs can be a list or separate arguments.

    e.g.::

        a = np.array([1, 2, 3])
        b = np.array([1, 2])

        XY = ndgrid(a, b)
            > [[1 1]
               [2 1]
               [3 1]
               [1 2]
               [2 2]
               [3 2]]

        X, Y = ndgrid(a, b, vector=False)
            > X = [[1 1]
                   [2 2]
                   [3 3]]
            > Y = [[1 2]
                   [1 2]
                   [1 2]]

    """

    # Read the keyword arguments, and only accept a vector=True/False
    vector = kwargs.pop('vector', True)
    assert isinstance(vector, bool), "'vector' keyword must be a bool"
    assert len(kwargs) == 0, "Only 'vector' keyword accepted"

    # you can either pass a list [x1, x2, x3] or each seperately
    if type(args[0]) == list:
        xin = args[0]
    else:
        xin = args

    # Each vector needs to be a numpy array
    assert np.all(
        [isinstance(x, np.ndarray) for x in xin]
    ), "All vectors must be numpy arrays."

    if len(xin) == 1:
        return xin[0]
    elif len(xin) == 2:
        XY = np.broadcast_arrays(mkvc(xin[1], 1), mkvc(xin[0], 2))
        if vector:
            X2, X1 = [mkvc(x) for x in XY]
            return np.c_[X1, X2]
        else:
            return XY[1], XY[0]
    elif len(xin) == 3:
        XYZ = np.broadcast_arrays(
            mkvc(xin[2], 1), mkvc(xin[1], 2), mkvc(xin[0], 3)
        )
        if vector:
            X3, X2, X1 = [mkvc(x) for x in XYZ]
            return np.c_[X1, X2, X3]
        else:
            return XYZ[2], XYZ[1], XYZ[0]


def check_xyz_dim(xyz, dim=3, dtype=float):
    """ Checks if the arguments are compatible with the expected dimensionality and type

    If xyz is a list or tuple of arrays, this will attempt to stack the arrays along the
    last dimension, ensuring all arrays are of consistent shapes with each other.

    Parameters
    ----------
    xyz : (dim,) tuple of numpy.ndarray, or (..., dim) numpy.ndarray
        The array to check
    dim : int, optional
        The expected dimensionality
    dtype : DataType, optional
        The requested data type for the array

    Returns
    -------
    xyz : (..., dim) numpy.ndarray of `dtype`
        The validated array
    """
    if isinstance(xyz, tuple):
        xyz = np.stack(xyz, axis=-1)
    xyz = np.asarray(xyz, dtype=dtype)
    if xyz.shape[-1] != dim:
        raise ValueError(
            f"Unexpected dimensionality of array, expected {dim}, saw {xyz.shape[-1]}"
        )
    return xyz



"""
oksar3

Program to calcuate forward models of interferograms, strain tensor, etc.
from Okada subroutine.

Heritage:
    - originally fringes.c written by Barry Parsons
    - updated to oksar                                 tjw
    - oksar_strain: added strain tensor calculation    tjw
    - oksar3:       added new line of sight calculator tjw feb 2003
    - Modified into Python by RowanCockett, 3point Science Aug 2014


"""

import numpy as np
import utm
import matplotlib.pyplot as plt
from datetime import datetime

def _date_time_from_json(value):
    if len(value) == 10:
        return datetime.strptime(value.replace('-', '/'),'%Y/%m/%d')
    return datetime.strptime(value, '%Y-%m-%dT%H:%M:%SZ')


class EarthquakeInterferogram:
    """Interferogram class"""

    def __init__(
        self,
        data,
        title,
        location,
        location_UTM_zone,
        shape,
        pixel_size,
        ref,
        ref_incidence,
        satellite_azimuth,
        satellite_altitude,
        processed_by,
        scaling=1.0,
        satellite_fringe_interval=0.028333,
        local_earth_radius=6371000.,
        local_rigidity=3e10,
        **kwargs
    ):

        self.data = data
        self.title = title
        self.location = location
        self.location_UTM_zone = location_UTM_zone
        self.shape = shape
        self.pixel_size = pixel_size
        self.ref = ref
        self.ref_incidence = ref_incidence
        self.satellite_azimuth = satellite_azimuth
        self.satellite_altitude = satellite_altitude
        self.processed_by = processed_by
        self.scaling = scaling
        self.satellite_fringe_interval = satellite_fringe_interval
        self.local_earth_radius = local_earth_radius
        self.local_rigidity = local_rigidity

        kwargs_list = [
            'description', 'event_country', 'event_name', 'copyright', 'data_source',
            'satellite_name', 'event_gcmt_id', 'date1', 'date2', 'processed_date', 'event_date'
        ]

        for k in kwargs_list:
            if k in kwargs.keys():
                setattr(self, k, kwargs.pop(k))

        super().__init__(**kwargs)

    @property
    def data(self):
        """Processed interferogram data (unwrapped).

        Returns
        -------
        numpy.ndarray of float
            Processed interferogram data (unwrapped)

        """
        return self._data

    @data.setter
    def data(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError("Location must be a array_like of float")

        self._data = var

    @property
    def title(self):
        """A title.

        Returns
        -------
        str
            A title.

        """
        return self._title

    @title.setter
    def title(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("Title is not a string or data type that can be converted to string")
        self._title = var

    @property
    def location(self):
        """Interferogram location (bottom N, left E).

        Returns
        -------
        (2,) numpy.ndarray
            Interferogram location (bottom N, left E)

        """
        return self._location

    @location.setter
    def location(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError(
                "Location must be array_like of float"
            )
        var = np.squeeze(var)
        if var.shape != (2, ):
            raise ValueError(
                f"Location must be of shape (2,) not {var.shape}"
            )

        self._location = var

    @property
    def location_UTM_zone(self):
        """UTM zone for the interferogram.

        Returns
        -------
        int
            UTM zone for the interferogram

        """
        return self._location_UTM_zone

    @location_UTM_zone.setter
    def location_UTM_zone(self, var):
        try:
            var = int(var)
        except:
            raise TypeError("Location_UTM_zone must be integer")

        self._location_UTM_zone = var

    @property
    def shape(self):
        """Shape of the interferogram defined by number of pixels in
        the North and Easting directions, respectively; i.e. (n_pix_N, n_pix_E).

        Returns
        -------
        (2,) tuple of int
            Interferogram shape, (number in Northing, number in Easting).

        """
        return self._shape

    @shape.setter
    def shape(self, var):
        try:
            var = tuple(int(v) for v in var)
        except:
            raise TypeError("shape must be a array_like of int")

        if len(var) != 2:
            raise ValueError(f"shape must be length 2, not{len(var)}")

        self._shape = var

    @property
    def pixel_size(self):
        """The Northing and Easting dimensions of each pixel.

        Returns
        -------
        (2,) numpy.ndarray of float
            The Northing and Easting dimensions of each pixel.

        """
        return self._pixel_size

    @pixel_size.setter
    def pixel_size(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError("pixel_size must be array_like of float")
        var = np.squeeze(var)
        if var.shape != (2,):
            raise TypeError(f"pixel_size must have shape (2, ) not {var.shape}")

        self._pixel_size = var

    @property
    def ref(self):
        """Interferogram reference location (bottom N, left E).

        Returns
        -------
        (2,) numpy.ndarray
            Interferogram reference location (bottom N, left E)

        """
        return self._ref

    @ref.setter
    def ref(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError(
                "ref must be a array_like of float"
            )
        var = np.squeeze(var)
        if var.shape != (2, ):
            raise ValueError(f"ref must have shape (2, ) not {var.shape}")

        self._ref = var

    @property
    def ref_incidence(self):
        """Incidence angle.

        Returns
        -------
        float
            Incidence angle

        """
        return self._ref_incidence

    @ref_incidence.setter
    def ref_incidence(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("ref_incidence must be a float")

        self._ref_incidence = var

    @property
    def scaling(self):
        """Interferogram scaling factor.

        Returns
        -------
        float
            Interferogram scaling factor. Default = 1.0

        """
        return self._scaling

    @scaling.setter
    def scaling(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("scaling must be a float")

        self._scaling = var

    @property
    def satellite_fringe_interval(self):
        """Satellite fringe interval.

        Returns
        -------
        float
            Satellite fringe interval

        """
        return self._satellite_fringe_interval

    @satellite_fringe_interval.setter
    def satellite_fringe_interval(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("satellite_fringe_interval must be a float")

        self._satellite_fringe_interval = var

    @property
    def satellite_azimuth(self):
        """Satellite azimuth.

        Returns
        -------
        float
            Satellite azimuth

        """
        return self._satellite_azimuth

    @satellite_azimuth.setter
    def satellite_azimuth(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("satellite_azimuth must be a float")

        self._satellite_azimuth = var

    @property
    def satellite_altitude(self):
        """Satellite altitude.

        Returns
        -------
        float
            Satellite altitude

        """
        return self._satellite_altitude

    @satellite_altitude.setter
    def satellite_altitude(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("satellite_altitude must be a float")

        self._satellite_altitude = var

    @property
    def local_earth_radius(self):
        """Local Earth radius.

        Returns
        -------
        float
            Local Earth radius. Default = 6371000.

        """
        return self._local_earth_radius

    @local_earth_radius.setter
    def local_earth_radius(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("local_earth_radius must be a float")

        self._local_earth_radius = var

    @property
    def local_rigidity(self):
        """Local rigidity.

        Returns
        -------
        float
            Local rigidity. Default = 3e10

        """
        return self._local_rigidity

    @local_rigidity.setter
    def local_rigidity(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("local_rigidity must be a float")

        self._rigidity = var

    @property
    def processed_by(self):
        """A string stating who processed the data.

        Returns
        -------
        str
            A string stating who processed the data.

        """
        return self._processed_by

    @processed_by.setter
    def processed_by(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("processed_by not a string or data type that can be converted to string")

        self._processed_by = var

    @property
    def description(self):
        """The description of the event.

        Returns
        -------
        str
            The description event. Default = 'My Earthquake'

        """
        return self._description

    @description.setter
    def description(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("description is not a string or data type that can be converted to string")
        self._description = var

    @property
    def satellite_name(self):
        """A string stating the name of the satellite.

        Returns
        -------
        str
            A string stating who the name of the satellite.

        """
        return self._satellite_name

    @satellite_name.setter
    def satellite_name(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("satellite_name not a string or data type that can be converted to string")

        self._satellite_name = var

    @property
    def copyright(self):
        """A string providing any copyright information.

        Returns
        -------
        str
            A string providing any copyright information.

        """
        return self._copyright

    @copyright.setter
    def copyright(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("copyright not a string or data type that can be converted to string")

        self._copyright = var

    @property
    def data_source(self):
        """The source of the data.

        Returns
        -------
        str
            The source of the data.

        """
        return self._data_source

    @data_source.setter
    def data_source(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("data_source not a string or data type that can be converted to string")

        self._data_source = var

    @property
    def event_gcmt_id(self):
        """The GCMT ID for the event.

        Returns
        -------
        str
            The GCMT ID for the event.

        """
        return self._event_gcmt_id

    @event_gcmt_id.setter
    def event_gcmt_id(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("event_gcmt_id not a string or data type that can be converted to string")

        self._event_gcmt_id = var

    @property
    def event_name(self):
        """The name of the event.

        Returns
        -------
        str
            The name of the event.

        """
        return self._event_name

    @event_name.setter
    def event_name(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("event_name not a string or data type that can be converted to string")

        self._event_name = var

    @property
    def event_country(self):
        """The country where the event occurred.

        Returns
        -------
        str
            The country where the event occurred.

        """
        return self._event_country

    @event_country.setter
    def event_country(self, var):
        try:
            var = str(var)
        except:
            raise TypeError("event_country not a string or data type that can be converted to string")

        self._event_country = var

    @property
    def date1(self):
        """Date 1

        Returns
        -------
        datetime.datetime
            The date and time
        """
        return self._date1

    @date1.setter
    def date1(self, date_time):
        if not isinstance(date_time, datetime):
            try:
                date_time = _date_time_from_json(date_time)
            except:
                raise TypeError(
                    "date_time must be instance of 'datetime.datetime' or a"
                    "json formatted date-time. Got {} instead".format(type(date_time))
                    )
        self._date1 = date_time

    @property
    def date2(self):
        """Date 2

        Returns
        -------
        datetime.datetime
            The date and time
        """
        return self._date2

    @date2.setter
    def date2(self, date_time):
        if not isinstance(date_time, datetime):
            try:
                date_time = _date_time_from_json(date_time)
            except:
                raise TypeError(
                    "date_time must be instance of 'datetime.datetime' or a"
                    "json formatted date-time. Got {} instead".format(type(date_time))
                    )
        self._date2 = date_time
    @property
    def processed_date(self):
        """Processed date

        Returns
        -------
        datetime.datetime
            The date and time
        """
        return self._processed_date

    @processed_date.setter
    def processed_date(self, date_time):
        if not isinstance(date_time, datetime):
            try:
                date_time = _date_time_from_json(date_time)
            except:
                raise TypeError(
                    "date_time must be instance of 'datetime.datetime' or a"
                    "json formatted date-time. Got {} instead".format(type(date_time))
                    )
        self._processed_date = date_time

    @property
    def event_date(self):
        """Event date

        Returns
        -------
        datetime.datetime
            The date and time
        """
        return self._processed_date

    @event_date.setter
    def event_date(self, date_time):
        if not isinstance(date_time, datetime):
            try:
                date_time = _date_time_from_json(date_time)
            except:
                raise TypeError(
                    "date_time must be instance of 'datetime.datetime' or a"
                    "json formatted date-time. Got {} instead".format(type(date_time))
                    )
        self._event_date = date_time

    def _get_plot_data(self):

        vectorNx = (
            np.r_[
                0,
                np.cumsum(
                    (self.pixel_size[0],) * self.shape[0]
                )
            ] + self.location[0]
        )
        vectorNy = (
            np.r_[
                0,
                np.cumsum(
                    (self.pixel_size[1],) * self.shape[1]
                )
            ] + self.location[1]
        ) - self.pixel_size[1] * self.shape[1]

        data = self.data.copy()
        data = np.flipud(data.reshape(self.shape, order='F').T)
        data[data == 0] = np.nan
        data *= self.scaling

        return vectorNx, vectorNy, data

    def plot_interferogram(self, wrap=True, ax=None):
        """Plot interferogram

        Parameters
        ----------
        wrap: bool
            If ``True``, wrap the function
        ax: matplotlib.ax.Axes
            An axes object

        Returns
        -------
        matplotlib.ax.pcolormesh
            The inteferogram plot

        """

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        vectorNx, vectorNy, data = self._get_plot_data()

        if wrap:
            cmap = plt.cm.hsv
            data = data % self.satellite_fringe_interval
            vmin, vmax = 0.0, self.satellite_fringe_interval
        else:
            cmap = plt.cm.jet
            vmin = np.nanmin(data)
            vmax = np.nanmax(data)

        out = ax.pcolormesh(
            vectorNx,
            vectorNy,
            np.ma.masked_where(np.isnan(data), data),
            vmin=vmin,
            vmax=vmax,
            cmap=cmap
        )

        ax.set_title(self.title)
        ax.axis('equal')
        ax.set_xlabel('Easting, m (UTM Zone {})'.format(
            self.location_UTM_zone
        ))
        ax.set_ylabel('Northing, m')

        cb = plt.colorbar(out, ax=ax)
        cb.set_label('Displacement, m')

        return out

    def plot_mask(self, ax=None, opacity=0.2):
        """Plot masked interferogram

        Parameters
        ----------
        ax: matplotlib.ax.Axes
            An axes object
        opacity: float
            The opacity, default = 0.2

        Returns
        -------
        matplotlib.ax.pcolormesh
            The masked inteferogram plot

        """

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        vectorNx, vectorNy, data = self._get_plot_data()

        from matplotlib import colors
        cmap = colors.ListedColormap([(1, 1, 1, opacity)])

        out = ax.pcolormesh(
            vectorNx,
            vectorNy,
            np.ma.masked_where(~np.isnan(data), data),
            cmap=cmap
        )

        ax.set_title(self.title)
        ax.axis('equal')
        ax.set_xlabel('Easting, m (UTM Zone {})'.format(
            self.location_UTM_zone
        ))
        ax.set_ylabel('Northing, m')

        return out

    def get_LOS_vector(self, locations):
        """calculate beta - the angle at earth center between reference point
        and satellite nadir

        Parameters
        ----------
        locations: list
            list of locations

        Returns
        -------
        numpy.ndarray
            The LOS vectors
        """
        if not isinstance(locations, list):
            locations = [locations]

        utmZone = self.location_UTM_zone
        refPoint = self.ref
        satAltitude = self.satellite_altitude
        satAzimuth = self.satellite_azimuth
        satIncidence = self.ref_incidence
        earthRadius = self.local_earth_radius

        DEG2RAD = np.pi / 180.
        alpha = satIncidence * DEG2RAD
        beta = (earthRadius / (satAltitude + earthRadius)) * np.sin(alpha)
        beta = alpha - np.arcsin(beta)
        beta = beta / DEG2RAD

        # calculate angular separation of (x,y) from satellite track passing
        # through (origx, origy) with azimuth satAzimuth

        # Long lat **NOT** lat long
        origy, origx = utm.to_latlon(
            refPoint[0], refPoint[1], np.abs(utmZone), northern=utmZone > 0
        )

        xy = np.array([
            utm.to_latlon(u[0], u[1], np.abs(utmZone), northern=utmZone > 0)
            for u in locations
        ])
        y = xy[:, 0]
        x = xy[:, 1]

        angdist = self._ang_to_gc(x, y, origx, origy, satAzimuth)

        # calculate beta2, the angle at earth center between roaming point and
        # satellite nadir track, assuming right-looking satellite

        beta2 = beta - angdist
        beta2 = beta2 * DEG2RAD

        # calculate alpha2, the new incidence angle

        alpha2 = np.sin(beta2) / (
            np.cos(beta2) - (earthRadius / (earthRadius + satAltitude))
        )
        alpha2 = np.arctan(alpha2)
        alpha2 = alpha2 / DEG2RAD

        # calculate pointing vector

        satIncidence = 90 - alpha2
        satAzimuth = 360 - satAzimuth

        los_x = -np.cos(satAzimuth * DEG2RAD) * np.cos(satIncidence * DEG2RAD)
        los_y = -np.sin(satAzimuth * DEG2RAD) * np.cos(satIncidence * DEG2RAD)
        los_z = np.sin(satIncidence * DEG2RAD)

        return np.squeeze(np.array([los_x, los_y, los_z]))

    @staticmethod
    def _ang_to_gc(x, y, origx, origy, satAzimuth):
        """
        Calculate angular distance to great circle passing through
        given point
        """

        Ngc = np.zeros(3)
        cartxy = np.zeros((len(x), 3))
        satAzimuth = np.deg2rad(satAzimuth)
        origx = np.deg2rad(origx)
        origy = np.deg2rad(origy)

        x = np.deg2rad(x)
        y = np.deg2rad(y)

        # 1. calc geocentric norm vec to great circle, Ngc = Rz*Ry*Rx*[0;1;0]
        #    where Rz = rotation of origx about geocentric z-axis
        #    where Ry = rotation of origy about geocentric y-axis
        #    where Rx = rotation of satAzimuth about geocentric x-axis
        #    and [0;1;0] is geocentric norm vec to N-S Great Circle through 0 0

        Ngc[0] = (
                    (np.sin(satAzimuth) * np.sin(origy) * np.cos(origx)) -
                    (np.cos(satAzimuth) * np.sin(origx))
                 )
        Ngc[1] = (
                    (np.sin(satAzimuth) * np.sin(origy) * np.sin(origx)) +
                    (np.cos(satAzimuth) * np.cos(origx))
                 )
        Ngc[2] = -np.sin(satAzimuth) * np.cos(origy)

        # 2. calculate unit vector geocentric coordinates for lon/lat
        #    position (x,y)

        cartxy[:, 0] = np.cos(x) * np.cos(y)
        cartxy[:, 1] = np.sin(x) * np.cos(y)
        cartxy[:, 2] = np.sin(y)

        # 3. Dot product between Ngc and cartxy gives angle 90 degrees
        #    bigger than what we want

        angdist = (
                        Ngc[0]*cartxy[:, 0] +
                        Ngc[1]*cartxy[:, 1] +
                        Ngc[2]*cartxy[:, 2]
                  )

        angdist = np.rad2deg(np.arccos(angdist)) - 90

        return angdist


class Oksar:

    def __init__(
        self,
        O,
        U,
        V,
        center,
        depth_top,
        depth_bottom=1e4,
        strike=0.,
        dip=45.,
        rake=90,
        slip=0.5,
        length=1e4,
        beta=3e10,
        mu=3e10,
        shape=(300, 300)
    ):

        self.O = O
        self.U = U
        self.V = V
        self.center = center
        self.depth_top = depth_top
        self.depth_bottom = depth_bottom
        self.strike = strike
        self.dip = dip
        self.rake = rake
        self.slip = slip
        self.length = length
        self.beta = beta
        self.mu = mu
        self.shape = shape

    @property
    def O(self):
        """Origin of the simulation domain.

        Returns
        -------
        (2,) numpy.ndarray
            Origin of the simulation domain.

        """
        return self._O

    @O.setter
    def O(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError("O must be array_like of float")
        var = np.squeeze(var)
        if var.shape != (2, ):
            raise ValueError(f"O must have a shape (2, ) not {var.shape}")
        self._O = var

    @property
    def U(self):
        """U direction of the simulation domain

        Returns
        -------
        (2,) numpy.ndarray
            U direction of the simulation domain

        """
        return self._U

    @U.setter
    def U(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError(
                "U must be a array_like of float"
            )
        var = np.squeeze(var)
        if var.shape != (2, ):
            raise ValueError(f"U must have a shape of (2, ) not {var.shape}")
        self._U = var

    @property
    def V(self):
        """V direction of the simulation domain

        Returns
        -------
        (2,) numpy.ndarray
            V direction of the simulation domain

        """
        return self._location

    @V.setter
    def V(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError("V must be array_like of float")
        var = np.squeeze(var)
        if var.shape != (2, ):
            raise ValueError(f"V must have a shape of (2, ) not {var.shape}")
        self._V = var


    @property
    def center(self):
        """Earthquake epicenter (bottom N, left E).

        Returns
        -------
        )2,) numpy.ndarray
            Earthquake epicenter (bottom N, left E)

        """
        return self._center

    @center.setter
    def center(self, var):
        try:
            var = np.asarray(var, dtype=float)
        except:
            raise TypeError("center must be array_like of float")
        var = np.squeeze(var)
        if var.shape != (2, ):
            raise ValueError(f"center must have a shape of (2, ) not {var.shape}")
        self._center = var

    @property
    def depth_top(self):
        """Depth to the top of the fault (m)

        Returns
        -------
        float
            Depth to the top of the fault (m)

        """
        return self._depth_top

    @depth_top.setter
    def depth_top(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("depth_top must be a float")

        if var < 0.:
            raise ValueError("depth_top must be equal or greather than 0")

        self._depth_top = var

    @property
    def depth_bottom(self):
        """Depth to the bottom of the fault (m)

        Returns
        -------
        float
            Depth to the bottom of the fault (m)

        """
        return self._depth_bottom

    @depth_bottom.setter
    def depth_bottom(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("depth_bottom must be a float")

        if var < self.depth_top:
            raise ValueError("depth_bottom must be larger than depth_top")

        self._depth_bottom = var

    @property
    def strike(self):
        """Strike angle (0 to 360) in degrees

        Returns
        -------
        float
            Strike angle (0 to 360)

        """
        return self._strike

    @strike.setter
    def strike(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("strike must be a float")

        if (var < 0.) | (var > 360.):
            raise ValueError("strike must be within [0, 360]")

        self._strike = var

    @property
    def dip(self):
        """Strike angle (0 to 90) in degrees

        Returns
        -------
        float
            Strike angle (0 to 90) in degrees

        """
        return self._dip

    @dip.setter
    def dip(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("dip must be a float")

        if (var < 0.) | (var > 90.):
            raise ValueError("dip must be within [0, 90]")

        self._dip = var

    @property
    def rake(self):
        """Rake angle (-180 to 180)

        Returns
        -------
        float
            Rake angle (-180 to 180)

        """
        return self._rake

    @rake.setter
    def rake(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("rake must be a float")

        if (var < -180) | (var > 180.):
            raise ValueError("rake must be within [-180, 180]")

        self._rake = var

    @property
    def slip(self):
        """Slip

        Returns
        -------
        float
            Slip

        """
        return self._slip

    @slip.setter
    def slip(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("slip must be a float")

        if var < 0.:
            raise ValueError("slip must be greater than or equal to 0")

        self._slip = var

    @property
    def length(self):
        """Length

        Returns
        -------
        float
            Length

        """
        return self._length

    @length.setter
    def length(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("length must be a float")

        if var < 0.:
            raise ValueError("length must be greater than or equal to 0")

        self._length = var

    @property
    def beta(self):
        """Beta parameter

        Returns
        -------
        float
            Beta parameter

        """
        return self._beta

    @beta.setter
    def beta(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("beta must be a float")

        self._beta = var

    @property
    def mu(self):
        """Mu parameter

        Returns
        -------
        float
            Mu parameter

        """
        return self._mu

    @mu.setter
    def mu(self, var):
        try:
            var = float(var)
        except:
            raise TypeError("mu must be a float")

        self._mu = var

    @property
    def shape(self):
        """Shape of the interferogram defined by number of pixels in
        the North and Easting directions, respectively; i.e. (n_pix_N, n_pix_E).

        Returns
        -------
        (2) numpy.ndarray of int
            Interferogram shape.

        """
        return self._shape

    @shape.setter
    def shape(self, var):
        try:
            var = np.array(var, dtype=int)
        except:
            raise TypeError("shape must be a (2) array_like of int")

        if len(var) != 2:
            raise TypeError("shape must be a (2) array_like of int")

        self._shape = var

    @property
    def simulation_grid(self):
        """Compute simulation grid

        Returns
        -------
        numpy.ndarray
            Simulation grid
        """
        shape = self.shape
        origin = self.O
        R = np.stack([self.U, self.V])
        square = np.meshgrid(
            np.linspace(0, 1, shape[0]),
            np.linspace(0, 1, shape[1])
        )
        square = np.stack([square[0].flatten(), square[1].flatten()], axis=-1)
        grid = origin + np.einsum('ij,...j->...i', R, square)

        return grid

    @property
    def displacement_vector(self):
        """Compute displacement vector

        Returns
        -------
        numpy.ndarray
            Displacement vector

        """

        vec = self.simulation_grid
        x, y = vec[:, 0], vec[:, 1]

        DEG2RAD = 0.017453292519943
        alpha = (self.beta + self.mu) / (self.beta + 2.0 * self.mu)

        #  Here we could loop over models
        flt_x = self.center[0]
        flt_y = self.center[1]
        strike = self.strike
        dip = self.dip
        rake = self.rake
        slip = self.slip
        length = self.length
        hmin = self.depth_top
        hmax = self.depth_bottom

        rrake = (rake+90.0)*DEG2RAD
        sindip = np.sin(dip*DEG2RAD)
        w = (hmax-hmin)/sindip
        ud = slip*np.cos(rrake)
        us = -slip*np.sin(rrake)
        halflen = length/2.0
        al2 = halflen
        al1 = -al2
        aw1 = hmin/sindip
        aw2 = hmax/sindip

        if(hmin == 0.0):
            hmin = 0.00001

        sstrike = (strike+90.0)*DEG2RAD

        ct = np.cos(sstrike)
        st = np.sin(sstrike)

        X = ct * (-flt_x + x) - st * (-flt_y + y)
        Y = ct * (-flt_y + y) + st * (-flt_x + x)

        u = self._dc3d3(alpha, X, Y, -dip, al1, al2, aw1, aw2, us, ud)

        UX = ct*u[...,0] + st*u[...,1]
        UY = -st*u[...,0] + ct*u[...,1]
        UZ = u[..., 2]

        return np.stack([UX, UY, UZ], axis=-1)

    def _dc3d3(self, alpha, X, Y, dip, al1, al2, aw1, aw2, disl1, disl2):
        F0 = 0.0
        F1 = 1.0
        F2 = 2.0
        PI2 = 6.283185307179586
        EPS = 1.0E-6

        u = np.array([F0, F0, F0])
        dub = np.array([F0, F0, F0])

        #  %%dccon0 subroutine
        #  Calculates medium and fault dip constants
        c0_alp3 = (F1 - alpha) / alpha
        #  PI2/360
        pl8 = 0.017453292519943
        c0_sd = np.sin(dip*pl8)
        c0_cd = np.cos(dip*pl8)

        if(np.abs(c0_cd) < EPS):
            c0_cd = F0
            if(c0_sd > F0):
                c0_sd = F1

            if(c0_sd < F0):
                c0_sd = -F1

        c0_cdcd = c0_cd * c0_cd
        c0_sdcd = c0_sd * c0_cd

        #  %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

        p = Y * c0_cd
        q = Y * c0_sd

        jxi = ((X - al1) * (X - al2)) <= F0  # BOOLEAN
        jet = ((p - aw1) * (p - aw2)) <= F0  # BOOLEAN

        for k in [1., 2.]:
            et = 0.0
            if(k == 1):
                et = p-aw1
            else:
                et = p-aw2

            for j in [1., 2.]:
                xi = 0.0
                if(j == 1):
                    xi = X-al1
                else:
                    xi = X-al2

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                # %%dccon2 subroutine
                # % calculates station geometry constants for finite source

                dc_max = np.max(np.abs(np.c_[xi, et, q]))

                # dc_max = max(np.abs(xi),max(np.abs(et),np.abs(q)))

                xi[(np.abs(xi/dc_max) < EPS) | (np.abs(xi) < EPS)] = F0

                et[(np.abs(et/dc_max) < EPS) | (np.abs(et) < EPS)] = F0

                q[(np.abs(q/dc_max) < EPS) | (np.abs(q) < EPS)] = F0

                dc_xi = xi
                dc_et = et
                dc_q = q
                c2_r = np.sqrt(dc_xi*dc_xi + dc_et*dc_et + dc_q*dc_q)

                if np.any(c2_r == F0):
                    raise Exception('singularity error ???')

                c2_y = dc_et * c0_cd + dc_q * c0_sd
                c2_d = dc_et * c0_sd - dc_q * c0_cd
                c2_tt = np.arctan(dc_xi * dc_et / (dc_q * c2_r))
                c2_tt[dc_q == F0] = F0

                rxi = c2_r + dc_xi
                c2_x11 = F1/(c2_r*rxi)
                c2_x11[(dc_xi < F0) & (dc_q == F0) & (dc_et == F0)] = F0

                ret = c2_r + dc_et
                if np.any(ret < 1e-14):
                    raise Exception('dccon2 b %f %f %f %f %f' % (
                        ret, c2_r, dc_et, dc_q, dc_xi
                    ))

                c2_ale = np.log(ret)
                c2_y11 = F1/(c2_r*ret)

                ind = (dc_et < F0) & (dc_q == F0) & (dc_xi == F0)

                # if((c2_r-dc_et) < 1e-14):
                #     raise Exception('dccon2 a %f %f %f %f %f' % (
                #         c2_3-dc_et, c2_r, dc_et, dc_q, dc_xi)
                #     )

                c2_ale[ind] = -np.log(c2_r[ind]-dc_et[ind])
                c2_y11[ind] = F0

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                if np.any(
                    (
                        (q == F0) &
                        (
                            ((jxi) & (et == F0)) |
                            ((jet) & (xi == F0))
                        )
                    ) | (c2_r == F0)
                ):
                    raise Exception('singular problems: 2')

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
                # ub subroutine
                # part B of displacement and strain at depth due to buried
                # faults in semi-infinite medium

                rd = c2_r + c2_d
                if np.any(rd < 1e-14):
                    raise Exception('ub %f %f %f %f %f %f' % (
                        rd, c2_r, c2_d, xi, et, q
                    ))

                ai3 = 0.0
                ai4 = 0.0
                if(c0_cd != F0):
                    # xx replaces x in original subroutine
                    xx = np.sqrt(xi*xi+q*q)
                    ai4 = F1/c0_cdcd * (xi/rd*c0_sdcd + F2*np.arctan(
                        (
                            et*(xx+q*c0_cd) +
                            xx*(c2_r+xx)*c0_sd
                        ) / (xi*(c2_r+xx)*c0_cd)
                    ))
                    ai4[xi == F0] = F0

                    ai3 = (c2_y*c0_cd/rd - c2_ale + c0_sd*np.log(rd)) / c0_cdcd
                else:
                    rd2 = rd*rd
                    ai3 = (et/rd + c2_y*q/rd2 - c2_ale) / F2
                    ai4 = xi*c2_y/rd2/F2

                ai1 = -xi/rd*c0_cd - ai4*c0_sd
                ai2 = np.log(rd) + ai3*c0_sd
                qx = q*c2_x11
                qy = q*c2_y11

                # strike-slip contribution
                if(disl1 != 0.0):
                    du2x = - xi*qy - c2_tt - c0_alp3 * ai1 * c0_sd
                    du2y = - q/c2_r + c0_alp3*c2_y/rd*c0_sd
                    du2z = q*qy - c0_alp3*ai2*c0_sd
                    du2 = np.array([du2x, du2y, du2z])
                    dub = du2 * (disl1 / PI2)
                else:
                    dub = np.zeros(3)

                # dip-slip contribution
                if(disl2 != F0):
                    du2x = - q/c2_r + c0_alp3 * ai3 * c0_sdcd
                    du2y = - et*qx - c2_tt - c0_alp3 * xi / rd * c0_sdcd
                    du2z = q*qx + c0_alp3 * ai4 * c0_sdcd
                    du2 = np.array([du2x, du2y, du2z])
                    dub = dub + (du2 * (disl2 / PI2))

                # %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

                dux = dub[0]
                duy = dub[1]*c0_cd - dub[2]*c0_sd
                duz = dub[1]*c0_sd + dub[2]*c0_cd
                du = np.array([dux, duy, duz])
                if((j+k) != 3):
                    u = + du + u
                else:
                    u = - du + u
        return u

    def plot_displacement(self, eq=None, ax=None, wrap=True, mask_opacity=0.2):
        """Plot displacement/

        Parameters
        ----------
        eq: EarthquakeInterferogram
            Instance of ``EarthquakeInterferogram``
        ax: matplotlib.ax.Axes
            Axes object
        wrap: bool
            If ``True``, wrap the function
        mask_opacity: float
            Masking opacity, default = 0.2

        Returns
        -------
        matplotlib.ax.pcolormesh
            Displacement plot

        """
        if eq is not None and not isinstance(eq, EarthquakeInterferogram):
            raise TypeError("eq must be an EarthquakeInterferogram")

        if ax is None:
            plt.figure()
            ax = plt.subplot(111)

        vectorNx = (
            np.r_[
                0,
                np.cumsum(
                    (self.U[0]/self.shape[0],) * self.shape[0]
                )
            ] + self.O[0]
        )
        vectorNy = (
            np.r_[
                0,
                np.cumsum(
                    (self.V[1]/self.shape[1],) * self.shape[1]
                )
            ] + self.O[1]
        )

        DIR = self.displacement_vector
        grid = self.simulation_grid
        if eq is None:
            LOS = np.array([0, 0, 1])
        else:
            LOS = eq.get_LOS_vector(grid)
        data = DIR.dot(LOS)
        data = np.flipud(data.reshape(self.shape, order='F').T)
        # data[data == 0] = np.nan
        # data *= self.scaling

        if wrap and eq is not None:
            cmap = plt.cm.hsv
            data = data % eq.satellite_fringe_interval
            vmin, vmax = 0.0, eq.satellite_fringe_interval
        else:
            cmap = plt.cm.jet
            vmin = np.nanmin(data)
            vmax = np.nanmax(data)

        out = ax.pcolormesh(
            vectorNx,
            vectorNy,
            np.ma.masked_where(np.isnan(data), data),
            vmin=vmin,
            vmax=vmax,
            cmap=cmap
        ),

        ax.axis('equal')
        ax.set_xlabel('Easting, m')
        ax.set_ylabel('Northing, m')

        cb = plt.colorbar(out[0], ax=ax)
        cb.set_label('Displacement, m')

        if eq is not None:
            mask = eq.plot_mask(ax=ax, opacity=mask_opacity)
            out = out[0], mask

        return out


def example():

    import requests
    dinar_file = requests.get(
        'https://storage.googleapis.com/simpeg/geoana/dinar.r4'
    )

    # data = np.fromstring(dinar_file.content, np.float32)

    # dinar = EarthquakeInterferogram(
    #     # uid='dinar',
    #     title='Dinar, Turkey',
    #     description=(
    #         'On October 1, 1995, a strong earthquake ruptured a section of '
    #         'the Dinar-Civril fault in SW Turkey. Around 30% of the buildings '
    #         'in the nearby town of Dinar were destroyed. 92 inhabitants were '
    #         'killed and over 200 injured.'
    #     ),
    #     event_country='Turkey',
    #     event_date='1995-09-30T18:00:00Z',
    #     event_gcmt_id='100195B',
    #     event_name='Dinar',
    #     copyright='ESA',
    #     data_source='ESA',
    #     date1='1995-08-12T18:00:00Z',
    #     date2='1995-12-31T17:00:00Z',
    #     processed_by='GarethFunning',
    #     processed_date='2003-01-20T17:00:00Z',
    #     ref_incidence=23,
    #     ref=[741140., 4230327.],
    #     scaling=0.0045040848895,
    #     local_earth_radius=6386232,
    #     local_rigidity=30000000000,
    #     location=[706216.0606, 4269238.9999],
    #     shape=(1024, 1024),
    #     location_UTM_zone=35,
    #     pixel_size=[80., 80.],
    #     satellite_altitude=788792,
    #     satellite_azimuth=192,
    #     satellite_fringe_interval=0.028333333,
    #     satellite_name='ERS',
    #     data=data
    # )

    data = np.fromstring(dinar_file.content, np.float32)
    title = 'Dinar, Turkey'
    location = [706216.0606, 4269238.9999]
    location_UTM_zone = 35
    shape = (1024, 1024)
    pixel_size = [80., 80.]
    ref= [741140., 4230327.]
    ref_incidence = 23
    satellite_azimuth=192
    satellite_altitude=788792
    processed_by='GarethFunning'

    dinar = EarthquakeInterferogram(
        data,
        title,
        location,
        location_UTM_zone,
        shape,
        pixel_size,
        ref,
        ref_incidence,
        satellite_azimuth,
        satellite_altitude,
        processed_by,
        scaling=0.0045040848895,
        satellite_fringe_interval=0.028333333,
        local_earth_radius=6386232,
        local_rigidity=30000000000,
        description=(
            'On October 1, 1995, a strong earthquake ruptured a section of '
            'the Dinar-Civril fault in SW Turkey. Around 30% of the buildings '
            'in the nearby town of Dinar were destroyed. 92 inhabitants were '
            'killed and over 200 injured.'
        ),
        event_country='Turkey',
        event_date='1995-09-30T18:00:00Z',
        event_gcmt_id='100195B',
        event_name='Dinar',
        copyright='ESA',
        data_source='ESA',
        date1='1995-08-12T18:00:00Z',
        date2='1995-12-31T17:00:00Z',
        processed_date='2003-01-20T17:00:00Z',
        satellite_name='ERS',
    )

    O = [706216.0606, 4187318.9999]
    U = [81920, 0]
    V = [0, 81920]
    center = [773728.2977967655, 4223586.816611591]
    depth_top=0

    dinar_fwd = Oksar(
        O,
        U,
        V,
        center,
        depth_top,
        depth_bottom=1.5e4,
        strike=329.6,
        dip=50.,
        rake=90,
        slip=0.5,
        length=11578.907244622129,
        beta=3e10,
        mu=3e10,
        shape=(300, 200)
    )

    return dinar, dinar_fwd



#!/usr/bin/env python
"""
geoana

Interactive geoscience (mostly) analytic functions.
"""

from distutils.core import setup
import sys

CLASSIFIERS = [
    'Development Status :: 4 - Beta',
    'Intended Audience :: Education',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Topic :: Scientific/Engineering',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Scientific/Engineering :: Physics',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: POSIX',
    'Operating System :: Unix',
    'Operating System :: MacOS',
    'Natural Language :: English',
]

def configuration(parent_package="", top_path=None):
    from numpy.distutils.misc_util import Configuration

    config = Configuration(None, parent_package, top_path)
    config.set_options(
        ignore_setup_xxx_py=True,
        assume_default_configuration=True,
        delegate_options_to_subpackages=True,
        quiet=True,
    )

    config.add_subpackage("geoana")

    return config

with open('README.rst') as f:
    LONG_DESCRIPTION = ''.join(f.readlines())

metadata = dict(
    name = 'geoana',
    version = '0.2.3',
    python_requires=">=3.6",
    setup_requires=[
        "numpy>=1.8",
        "cython>=0.2",
    ],
    install_requires = [
        'numpy>=1.8',
        'scipy>=0.13',
        'matplotlib',
        'utm',
        'empymod'
    ],
    author = 'SimPEG developers',
    author_email = 'lindseyheagy@gmail.com',
    description = 'Analytic expressions for geophysical responses',
    long_description = LONG_DESCRIPTION,
    keywords = 'geophysics, electromagnetics',
    url = 'https://www.simpeg.xyz',
    download_url = 'https://github.com/simpeg/geoana',
    classifiers=CLASSIFIERS,
    platforms = ['Windows', 'Linux', 'Solaris', 'Mac OS-X', 'Unix'],
    license='MIT License'
)

if len(sys.argv) >= 2 and (
    "--help" in sys.argv[1:]
    or sys.argv[1] in ("--help-commands", "egg_info", "--version", "clean")
):
    # For these actions, NumPy is not required.
    #
    # They are required to succeed without Numpy, for example when
    # pip is used to install discretize when Numpy is not yet present in
    # the system.
    try:
        from setuptools import setup
    except ImportError:
        from distutils.core import setup
else:
    if (len(sys.argv) >= 2 and sys.argv[1] in ("bdist_wheel", "bdist_egg")) or (
        "develop" in sys.argv
    ):
        # bdist_wheel/bdist_egg needs setuptools
        import setuptools

    from numpy.distutils.core import setup

    # Add the configuration to the setup dict when building
    # after numpy is installed
    metadata["configuration"] = configuration


setup(**metadata)
